import std/[json, options, os, paths, strutils, times, unicode]
import cliutils, docopt, illwill, mpv

import wdiwtlt/[cliconstants, commandbuffer, logging, medialibrary, models,
                scrolltext]

type
  CliMode {.pure.} = enum Direct, Command

  ScrollingDisplays* = ref object
    playing: ScrollText
    status: ScrollText
    statusDismissal: Option[DateTime]

  History = ref object
    entries: seq[string]
    idx: int

  RenderDimensions = ref object
    width, height: int
    playing: tuple[x, y: int]
    status: tuple[x, y: int]
    input: tuple[x, y: int]
    logs: tuple[x, y: int]

  CliContext = ref object
    cfg: CombinedConfig
    cmd: CommandBuffer
    cmdVisibleIdx: int              # first visible character in the cmd buffer
    curMediaFile: Option[MediaFile]
    d: ScrollingDisplays
    dims: RenderDimensions
    frame: int
    lib: MediaLibrary
    log: Option[Logger]
    mode: CliMode
    mpv: ptr handle
    msgDuration: Duration
    stop*: bool
    tb: TerminalBuffer

  CursorType {.pure.} = enum
    Block, BlockBlink, BarBlink, Underline

var ctx {.threadvar.}: CliContext


proc initMpv(): ptr handle =
  result = mpv.create()

  if result.isNil:
    raise newException(ValueError, "failed creating mpv context")

  # Enable default key bindings, so the user can actually interact with
  # the player (and e.g. close the window).
  result.set_option("terminal", "no")
  result.set_option("input-default-bindings", "yes")
  result.set_option("input-vo-keyboard", "yes")
  result.set_option("osc", false)
  #result.set_property("volume", "50.0")
  discard result.request_log_messages("no")

  # Done setting up options.
  check_error result.initialize()


proc cleanup() =
  terminate_destroy(ctx.mpv)
  illWillDeinit()
  echo ""


proc exitHandlerHook() {.noconv.} =
  # This is called when the user presses Ctrl+C
  # or when the program exits normally.
  echo "Ctrl-C"
  terminate_destroy(ctx.mpv)
  illWillDeinit()
  quit(QuitSuccess)


proc clearLine(ctx: var CliContext, y: int) =
  ctx.tb.write(0, y, ' '.repeat(ctx.tb.width))


proc writeAt(ctx: var CliContext, pos: tuple[x, y: int], text: string) =
  ctx.tb.write(pos.x, pos.y, text)


proc nextTick(ctx: var CliContext) =
  if ctx.d.statusDismissal.isSome and ctx.d.statusDismissal.get < now():
    ctx.d.status.text = ""
    ctx.d.statusDismissal = none[DateTime]()
    ctx.clearLine(ctx.dims.status.y)

  discard ctx.d.playing.nextTick
  discard ctx.d.status.nextTick


proc setMode(ctx: var CliContext, mode: CliMode) =
  ctx.mode = mode
  case mode
  of CliMode.Direct:
    hideCursor()
    ctx.tb.clear()
  of CliMode.Command:
    showCursor()
    ctx.cmd.mode = EditMode.Insert


proc initContext(args: Table[string, Value]): CliContext =
  var wdiwtltCfgFilename: string

  let log = getLogger("wdiwtlt")

  let cfgLocations =
    if args["--config"]: @[$args["--config"]]
    elif existsEnv("XDG_CONFIG_HOME"): @[getEnv("XDG_CONFIG_HOME") / "wdiwtlt.json"]
    else: @[$getEnv("HOME") / ".config" / "wdiwtlt.json" ]

  try: wdiwtltCfgFilename = findConfigFile(".wdiwtlt.json", cfgLocations)
  except ValueError:
    log.error "could not find wdiwtlt config file: " & wdiwtltCfgFilename
    if not args["--config"]:
      wdiwtltCfgFilename = cfgLocations[0]

      var cfgFile: File
      try:
        cfgFile = open(wdiwtltCfgFilename, fmWrite)
        cfgFile.write((%*{
          "libraryRoot":
            if args["--library-root"]:
              $args["--library-root"]
            else:
              getEnv("XDG_MUSIC_DIR", getEnv("HOME") / "music"),
          "dbPath": getEnv("XDG_DATA_HOME", (getEnv("HOME") / ".local/share")) / "wdiwtlt/db.sqlite",
        }).pretty)
        log.info "created sample config file at " & wdiwtltCfgFilename
      except CatchableError:
        log.error "could not write default .wdiwtlt.json to " &
          wdiwtltCfgFilename
      finally: close(cfgFile)

    quit(QuitFailure)

  log.debug("loading config from '$#'" % [wdiwtltCfgFilename])
  let cfg = initCombinedConfig(wdiwtltCfgFilename, args)

  let width = min(terminalWidth(), cfg.getVal("maxWidth", "80").parseInt)
  let height = min(terminalHeight(), cfg.getVal("maxHeight", "60").parseInt)

  result = CliContext(
    cfg: cfg,
    cmd: initCommandBuffer(),
    cmdVisibleIdx: 0,
    curMediaFile: none(MediaFile),
    d: ScrollingDisplays(
      playing: initScrollText("Nothing playing", width),
      status: initScrollText("Idle", width),
      statusDismissal: none[DateTime]()),
    dims: RenderDimensions(
      width: width,
      height: height,
      playing: (0, 0),
      status: (0, 1),
      input: (0, 2),
      logs: (0, 3)),
    frame: 0,
    lib: initMediaLibrary(
      rootDir = getVal(cfg, "libraryRoot").Path,
      dbPath = getVal(cfg, "dbPath").Path),
    log: log,
    mode: CliMode.Direct,
    mpv: initMpv(),
    msgDuration: initDuration(seconds = 5, milliseconds = 0),
    stop: false,
    tb: newTerminalBuffer(width, height))

  if logService.isSome:
    let customLogAppender = initCustomLogAppender(doLogMessage =
      proc (msg: LogMessage): void =
        ctx.writeAt(ctx.dims.logs, msg.message))

    #logService.get.clearAppenders()
    logService.get.addAppender(customLogAppender)


proc setCursorType(ct: CursorType) =
  case ct
  of CursorType.Block:      stdout.write("\x1b[2 q")
  of CursorType.BlockBlink: stdout.write("\x1b[1 q")
  of CursorType.BarBlink:   stdout.write("\x1b[5 q")
  of CursorType.Underline:  stdout.write("\x1b[4 q")


proc setCursorPosition(x, y: int) =
  stdout.write("\x1b[$#;$#H" % [$(y + 1), $(x + 1)])


proc msg(ctx: var CliContext, msg: string) =
  ctx.d.status.text = msg
  if msg.len > ctx.dims.width:
    ctx.d.statusDismissal = some(now() + ctx.msgDuration +
    initDuration(milliseconds = (msg.len - ctx.dims.width) * 192))
  else:
    ctx.d.statusDismissal = some(now() + ctx.msgDuration)


proc render(ctx: var CliContext) =
  # text moves at 5 characters per second (192ms per character)
  if ctx.frame mod 12 == 0: nextTick(ctx)

  ctx.clearLine(ctx.dims.playing.y)
  ctx.writeAt(ctx.dims.playing, ctx.d.playing)

  ctx.clearLine(ctx.dims.status.y)
  ctx.writeAt(ctx.dims.status, ctx.d.status)

  if ctx.mode == CliMode.Command:
    case ctx.cmd.mode:
    of EditMode.Insert:     setCursorType(CursorType.BarBlink)
    of EditMode.Overwrite:  setCursorType(CursorType.BlockBlink)
    else:                   setCursorType(CursorType.Block)

    ctx.clearLine(ctx.dims.input.y)

    var inputLine = $ctx.cmd
    let spaceAvailable = ctx.dims.width - 2
    # Check to see that the line entered fits entirely within the space
    # available to display it. We need two characters of space: one for the
    # colon and one for the cursor. If we don't have enough space, we need to
    # decide which portion of the line to display.
    if len(inputLine) >= spaceAvailable:
      # If the user has moved the cursor left of our visible index, follow them
      # back
      if ctx.cmd.idx < ctx.cmdVisibleIdx: ctx.cmdVisibleIdx = ctx.cmd.idx

      # If the user has moved right, push our index to the right to follow them
      elif (ctx.cmd.idx - ctx.cmdVisibleIdx) > spaceAvailable:
        ctx.cmdVisibleIdx = ctx.cmd.idx - spaceAvailable

      # If the user has started backspacing (we are showing less that we
      # could), follow them back
      elif (len(inputLine) - ctx.cmdVisibleIdx) < spaceAvailable:
        ctx.cmdVisibleIdx = len(inputLine) - spaceAvailable

      # Show the portion of the line starting at the visible index and ending
      # either with the end of the line or the end of the visible space
      # (whichever comes first). We need to check both because the user may
      # delete characters from the end of the line, which would make the
      # portion of the line we're currently showing shorter than the visible
      # space (don't try to access an index past the end of the string).
      inputLine = inputLine[
        ctx.cmdVisibleIdx ..<
        min(ctx.cmdVisibleIdx + spaceAvailable, len(inputLine))]

    elif ctx.cmdVisibleIdx > 0:
      # We know the line fits within the space available, but we're still
      # showing a subset of the line (probably because it was longer and the
      # user backspaced). Let's just reset and show the whole line now that we
      # can.
      ctx.cmdVisibleIdx = 0

    ctx.writeAt(ctx.dims.input, ":" & inputLine)
    setCursorPosition(
      ctx.dims.input.x + (ctx.cmd.idx - ctx.cmdVisibleIdx) + 1,
      ctx.dims.input.y)

  else:
    ctx.clearLine(ctx.dims.input.y)
    ctx.writeAt(ctx.dims.input, "WDIWTLT v" & VERSION)

  ctx.tb.display()


proc scanLibrary(ctx: var CliContext) =
  let counts = ctx.lib.rescanLibrary()


proc processCommand(ctx: var CliContext, command: string) =
  let parts = command.strip.split(' ', maxSplit=1)

  case parts[0]
  of "scan":                scanLibrary(ctx)
  of "q", "quit", "exit":   ctx.stop = true
  else: msg(ctx, "Unrecognized command: " & command)

proc mainLoop(ctx: var CliContext) =

  hideCursor()

  while not ctx.stop:
    let key = getKey()

    if ctx.mode == CliMode.Direct:
      case key
      of Key.Q:     ctx.stop = true
      of Key.Colon, Key.I: ctx.setMode(CliMode.Command)
      else: discard

    elif ctx.mode == CliMode.Command:
      case key
      of Key.Enter:
        let command = $ctx.cmd
        ctx.cmd.clear
        ctx.cmdVisibleIdx = 0
        ctx.cmd.mode = EditMode.Insert
        processCommand(ctx, command)
      of Key.Backspace: ctx.cmd.handleInput(Key.Backspace)
      of Key.Escape:
        if ctx.cmd.mode == EditMode.Command:
          ctx.setMode(CliMode.Direct)
          ctx.cmd.clear
          ctx.cmdVisibleIdx = 0
        else: ctx.cmd.handleInput(Key.Escape)
      else: ctx.cmd.handleInput(key)

    render(ctx)

    # target 60 FPS
    sleep(16)
    ctx.frame = (ctx.frame + 1) mod 1920


when isMainModule:

  discard enableLogging()
  try:
    let args = docopt(USAGE, version=VERSION)

    illWillInit(fullscreen=true)
    ctx = initContext(args)

    mainLoop(ctx)

  except Exception:
    let ex = getCurrentException()
    getLogger("wdiwtlt").error(%*{
      "msg": "Unhandled exception",
      "error": ex.msg,
      "trace": ex.getStackTrace()
    })
    cleanup()
    quit(QuitFailure)
  finally:
    cleanup()
