import std/[options]

import illwill

type
  EditMode* {.pure.} = enum Command, Insert, Overwrite, Visual

  Command = ref object
    buffer: string
    idx: int
    selectionStartIdx: Option[int]

  CommandBuffer* = ref object
    history: seq[Command]
    idx: int
    mode*: EditMode

func clamp(value, min, max: int): int =
  if value < min:
    return min
  elif value > max:
    return max
  else:
    return value

func initCommandBuffer*(): CommandBuffer =
  result = CommandBuffer(
    history: @[Command(
      buffer: "",
      idx: 0,
      selectionStartIdx: none(int))],
    idx: 0,
    mode: Insert)

proc cur(cb: CommandBuffer): Command = cb.history[cb.idx]
proc cur(cb: var CommandBuffer): Command = cb.history[cb.idx]

proc prev*(cb: var CommandBuffer) =
  cb.idx = clamp( cb.idx - 1, 0, cb.history.len - 1)

proc next*(cb: var CommandBuffer) =
  cb.idx = clamp( cb.idx + 1, 0, cb.history.len - 1)

proc insert*(cb: var CommandBuffer, s: string) =
  var cmd = cb.history[cb.idx]
  cmd.buffer = cmd.buffer[0..<cmd.idx] & s & cmd.buffer[cmd.idx..^1]
  cmd.idx += s.len

proc overwrite*(cb: var CommandBuffer, s: string) =
  var cmd = cb.history[cb.idx]
  if cmd.idx + s.len > cmd.buffer.len:
    cmd.buffer = cmd.buffer[0..<cmd.idx] & s
  else:
    cmd.buffer = cmd.buffer[0..<cmd.idx] & s & cmd.buffer[cmd.idx + s.len..^1]
  cmd.idx += s.len

proc write*(cb: var CommandBuffer, s: string) =
  if cb.mode == EditMode.Insert: cb.insert(s)
  elif cb.mode == EditMode.Overwrite: cb.overwrite(s)

proc delete*(cb: var CommandBuffer, backspace = true) =
  var cmd = cb.history[cb.idx]
  if backspace and cmd.idx > 0:
    cmd.buffer = cmd.buffer[0..<cmd.idx - 1] & cmd.buffer[cmd.idx..^1]
    cmd.idx -= 1
  elif not backspace and cmd.idx < cmd.buffer.len:
    cmd.buffer = cmd.buffer[0..<cmd.idx] & cmd.buffer[cmd.idx + 1..^1]

proc clear*(cb: var CommandBuffer) =
  cb.history.add(Command(
    buffer: "",
    idx: 0,
    selectionStartIdx: none(int)))
  cb.idx = cb.history.len - 1

proc left*(cb: var CommandBuffer) =
  cb.cur.idx = clamp(cb.cur.idx - 1, 0, cb.cur.buffer.len)

proc right*(cb: var CommandBuffer) =
  cb.cur.idx = clamp(cb.cur.idx + 1, 0, cb.cur.buffer.len)

proc toHome*(cb: var CommandBuffer) = cb.cur.idx = 0
proc toEnd*(cb: var CommandBuffer) = cb.cur.idx = cb.cur.buffer.len

proc backWord*(cb: var CommandBuffer) =
  var cmd = cb.history[cb.idx]
  while cmd.idx > 0 and cmd.buffer[cmd.idx - 1] == ' ': cmd.idx -= 1
  while cmd.idx > 0 and cmd.buffer[cmd.idx - 1] != ' ': cmd.idx -= 1

proc forwardWord*(cb: var CommandBuffer) =
  var cmd = cb.history[cb.idx]
  while cmd.idx < cmd.buffer.len and cmd.buffer[cmd.idx] != ' ': cmd.idx += 1
  while cmd.idx < cmd.buffer.len and cmd.buffer[cmd.idx] == ' ': cmd.idx += 1

proc endOfWord*(cb: var CommandBuffer) =
  var cmd = cb.history[cb.idx]
  cmd.idx = clamp(cmd.idx + 1, 0, cmd.buffer.len - 1)
  while cmd.idx < cmd.buffer.len and cmd.buffer[cmd.idx] == ' ': cmd.idx += 1
  while cmd.idx < cmd.buffer.len - 1 and
        cmd.buffer[cmd.idx + 1] != ' ':
    cmd.idx += 1

func toChar(k: Key): char = cast[char](ord(k))

proc handleInput*(cb: var CommandBuffer, key: Key) =
  case cb.mode
  of EditMode.Insert, EditMode.Overwrite:
    case key
    of Key.Escape: cb.mode = EditMode.Command
    of Key.Backspace: cb.delete()
    of Key.Delete: cb.delete(backspace = false)
    of Key.Left: cb.left()
    of Key.Right: cb.right()
    of Key.Up: cb.prev()
    of Key.Down: cb.next()
    of Key.Home: cb.toHome()
    of Key.End: cb.toEnd()
    of Key.CtrlH: cb.backWord()
    of Key.CtrlL: cb.forwardWord()
    elif key >= Key.Space and key <= Key.Tilde: cb.write($toChar(key))
    else: discard

  of EditMode.Command:
    case key
    of Key.Backspace: cb.delete()
    of Key.Delete,  Key.X: cb.delete(backspace = false)
    of Key.Left,    Key.H: cb.left()
    of Key.Right,   Key.L: cb.right()
    of Key.Up,      Key.J: cb.prev()
    of Key.Down,    Key.K: cb.next()
    of Key.Home,    Key.Zero:   cb.toHome()
    of Key.End,     Key.Dollar: cb.toEnd()
    of Key.B: cb.backWord()
    of Key.W: cb.forwardWord()
    of Key.E: cb.endOfWord()
    of Key.I: cb.mode = EditMode.Insert
    of Key.A:
      cb.cur.idx = clamp(cb.cur.idx + 1, 0, cb.cur.buffer.len)
      cb.mode = EditMode.Insert
    of Key.ShiftA:
      cb.cur.idx = cb.cur.buffer.len
      cb.mode = EditMode.Insert
    of Key.ShiftI:
      cb.cur.idx = 0
      cb.mode = EditMode.Insert
    of Key.ShiftR: cb.mode = EditMode.Overwrite
    of Key.V:
      cb.mode = EditMode.Visual
      cb.cur.selectionStartIdx = some(cb.cur.idx)
    else: discard

  of EditMode.Visual:
    case key
    of Key.Escape:
      cb.mode = EditMode.Command
      cb.cur.selectionStartIdx = none(int)
    of Key.Backspace: cb.left()
    of Key.Left,   Key.H: cb.left()
    of Key.Right,  Key.L: cb.right()
    of Key.Up,     Key.J:
      cb.cur.selectionStartIdx = none[int]()
      cb.prev()
    of Key.Down,   Key.K:
      cb.cur.selectionStartIdx = none[int]()
      cb.next()
    of Key.Home,   Key.Zero: cb.toHome()
    of Key.End,    Key.Dollar: cb.toEnd()
    of Key.B: cb.backWord()
    of Key.W: cb.forwardWord()
    of Key.V:
      cb.mode = EditMode.Command
      cb.cur.selectionStartIdx = none(int)
    else: discard


func `$`*(cb: CommandBuffer): string = cb.cur.buffer

func idx*(cb: CommandBuffer): int = cb.cur.idx
