import std/[dirs, files, json, options, paths, sequtils, strutils, times]
import db_connector/db_sqlite
import waterpark/sqlite
import fiber_orm, timeutils, uuids

export fiber_orm.NotFoundError
export fiber_orm.PaginationParams
export fiber_orm.PagedRecords
export fiber_orm.enableDbLogging
export sqlite.close

import ./[logging, models]

const schemaDDL = readFile("src/main/sql/media-library-schema.sql")
  .split(";")
  .filterIt(not isEmptyOrWhitespace(it))

type
  WdiwtltDb* = SqlitePool

func toJsonHook*(dt: DateTime): JsonNode = %(dt.formatIso8601)
proc fromJsonHook*(dt: var DateTime, n: JsonNode): void =
  dt = n.getStr.parseIso8601


proc createTables*(db: WdiwtltDb) =
  ## Create the database tables if they don't exist.

  db.withConnection conn:
    let rows = conn.getRow(sql("SELECT name FROM sqlite_master WHERE type = 'table' AND name = 'artists'"))

    if rows.len == 0 or isEmptyOrWhitespace(rows[0]):
      for ddl in schemaDDL:
        try: conn.exec(sql(ddl))
        except DbError:
          let ex = getCurrentException()
          getLogger("wdiwtlt/db").error(%*{
            "msg": "Failed creating initial schema.",
            "ddl": ddl,
            "error": ex.msg})
          raise ex

proc initDB*(dbPath: Path): SqlitePool =
  # Create the DB file if it doesn't exist.
  if not fileExists(dbPath):
    let dbDir = dbPath.splitFile().dir
    if not dirExists(dbDir):
      createDir(dbDir)

  result = newSqlitePool(10, dbPath.string)

  # Create the database tables if they don't exist.
  result.createTables()


generateProcsForModels(WdiwtltDb,
  [Artist, Album, MediaFile, Tag, Playlist, Bookmark, Image])

generateJoinTableProcs(WdiwtltDb, Artist, Album, "artists_albums")
generateJoinTableProcs(WdiwtltDb, Artist, MediaFile, "artists_media_files")
generateJoinTableProcs(WdiwtltDb, Album, MediaFile, "albums_media_files")
generateJoinTableProcs(WdiwtltDb, Playlist, MediaFile, "playlists_media_files")
generateJoinTableProcs(WdiwtltDb, MediaFile, Tag, "media_files_tags")
generateJoinTableProcs(WdiwtltDb, Artist, Image, "artists_images")
generateJoinTableProcs(WdiwtltDb, Album, Image, "albums_images")

generateLookup(WdiwtltDb, Artist, @["name"])
generateLookup(WdiwtltDb, Album, @["artistId"])
generateLookup(WdiwtltDb, Album, @["name"])
generateLookup(WdiwtltDb, Album, @["name", "year"])
generateLookup(WdiwtltDb, MediaFile, @["filePath"])
generateLookup(WdiwtltDb, MediaFile, @["fileHash"])

proc findAlbumsByArtistAndName*(
    db: WdiwtltDb,
    artistId: UUID,
    albumName: string,
    page: Option[PaginationParams] = none[PaginationParams]()): PagedRecords[Album] =
  ## Find albums by artist and name.

  db.withConnection conn:
    var query = """
      SELECT """ & columnNamesForModel(Album).join(",") & """
      FROM albums al
      JOIN artists_albums aral ON al.id = aral.album_id
      WHERE aral.artist_id = ? AND al.name = ? """

    let countStmt = """
      SELECT COUNT(*)
      FROM albums al
      JOIN artists_albums aral ON al.id = aral.album_id
      WHERE aral.artist_id = ? AND al.name = ? """

    if page.isSome: query &= getPagingClause(page.get)

    logQuery("findAlbumsByArtistAndName", query,
      [("artist_id", $artistId), ("album.name", albumName)])

    let values = @[$artistId, albumName]
    let records = conn.getAllRows(sql(query), values).mapIt(rowToModel(Album, it))

    result = PagedRecords[Album](
      pagination: page,
      records: records,
      totalRecords:
        if page.isNone: records.len
        else: conn.getRow(sql(countStmt), values)[0].parseInt)


proc findAlbumsByArtistAndName*(
    db: WdiwtltDb,
    artist: Artist,
    albumName: string,
    page: Option[PaginationParams] = none[PaginationParams]()): PagedRecords[Album] =
  ## Find albums by artist and name.
  return db.findAlbumsByArtistAndName(artist.id, albumName, page)


proc findAlbumsByArtistNameAndYear*(
    db: WdiwtltDb,
    artistId: UUID,
    albumName: string,
    year: int,
    page: Option[PaginationParams] = none[PaginationParams]()): PagedRecords[Album] =
  ## Find albums by artist name and year.

  db.withConnection conn:
    var query = """
      SELECT """ & columnNamesForModel(Album).join(",") & """
      FROM albums al
      JOIN artists_albums aral ON al.id = aral.album_id
      WHERE aral.artist_id = ? AND al.name = ? AND al.year = ? """

    let countStmt = """
      SELECT COUNT(*)
      FROM albums al
      JOIN artists_albums aral ON al.id = aral.album_id
      WHERE aral.artist_id = ? AND al.name = ? AND al.year = ? """

    if page.isSome: query &= getPagingClause(page.get)

    logQuery("findAlbumsByArtistNameAndYear", query,
      [("artist_id", $artistId), ("album.name", albumName), ("album.year", $year)])

    let values = @[$artistId, albumName, $year]
    let records = conn.getAllRows(sql(query), values).mapIt(rowToModel(Album, it))

    result = PagedRecords[Album](
      pagination: page,
      records: records,
      totalRecords:
        if page.isNone: records.len
        else: conn.getRow(sql(countStmt), values)[0].parseInt)

proc findAlbumsByArtistNameAndYear*(
    db: WdiwtltDb,
    artist: Artist,
    albumName: string,
    year: int,
    page: Option[PaginationParams] = none[PaginationParams]()): PagedRecords[Album] =
  ## Find albums by artist name and year.
  return db.findAlbumsByArtistNameAndYear(artist.id, albumName, year, page)

proc removeEmptyAlbums*(db: WdiwtltDb) =
  ## Remove albums that have no media files.

  db.withConnection conn:
    let query = """
        DELETE FROM albums WHERE id IN (
            SELECT DISTINCT al.id
            FROM albums al LEFT OUTER JOIN albums_media_files almf ON
                al.id = almf.album_id
            WHERE almf.album_id IS NULL)"""

    getLogger("wdiwtlt/db").debug(%*{
      "msg": "Deleting empty albums.",
      "query": query})

    conn.exec(sql(query))

proc removeEmptyArtists*(db: WdiwtltDb) =
  ## Remove artists that have no albums.

  db.withConnection conn:
    let query = """
        DELETE FROM artists WHERE id IN (
            SELECT DISTINCT ar.id
            FROM artists ar LEFT OUTER JOIN artists_albums aral ON
                ar.id = aral.artist_id
            WHERE aral.artist_id IS NULL)"""

    getLogger("wdiwtlt/db").debug(%*{
      "msg": "Deleting empty artists.",
      "query": query})

    conn.exec(sql(query))

proc removeEmptyPlaylists*(db: WdiwtltDb) =
  ## Remove playlists that have no media files.

  db.withConnection conn:
    let query = """
        DELETE FROM playlists WHERE id IN (
            SELECT DISTINCT pl.id
            FROM playlists pl LEFT OUTER JOIN playlists_media_files plmf ON
                pl.id = plmf.playlist_id
            WHERE plmf.playlist_id IS NULL)"""

    getLogger("wdiwtlt/db").debug(%*{
      "msg": "Deleting empty playlists.",
      "query": query})

    conn.exec(sql(query))
