import std/[dirs, files, json, options, paths, sequtils, strutils, sugar, times]
import namespaced_logging, timeutils, uuids

import ./[db, incremental_md5, logging, models, taglib]

type
  MediaLibrary* = ref object
    db: WdiwtltDb
    root: Path

proc initMediaLibrary*(rootDir: Path, dbPath: Path): MediaLibrary =
  ## Initialize the media library.

  let db = initDB(dbPath)

  return MediaLibrary(db: db, root: rootDir)

proc clean*(lib: MediaLibrary) =
  removeEmptyAlbums(lib.db)
  removeEmptyArtists(lib.db)
  removeEmptyPlaylists(lib.db)

  let expirationDate = now() - weeks(1)

  let expiredPlaylists = lib.db.findPlaylistsWhere(
    "user_created = false AND last_used_at < ?",
    [expirationDate.formatIso8601])

  let expiredBookmarks = lib.db.findBookmarksWhere(
    "user_created = false AND last_used_at < ?",
    [expirationDate.formatIso8601])

  for p in expiredPlaylists.records: discard lib.db.deletePlaylist(p.id)
  for b in expiredBookmarks.records: discard lib.db.deleteBookmark(b.id)

proc findOrCreateArtist*(lib: MediaLibrary, name: string): Artist =
  ## Find or create an artist record.

  let existing = lib.db.findArtistsByName(name)
  if existing.records.len > 0:
    return existing.records[0]

  getLogger("wdiwtlt/medialibrary").debug(%*{
    "msg": "Creating missing Artist record.",
    "method": "findOrCreateArtist",
    "artistName": name })
  return lib.db.createArtist(Artist(id: genUuid(), name: name))


proc findOrCreateAlbum*(
    lib: MediaLibrary,
    name: string,
    artists: seq[Artist],
    year: Option[int]): Album =

  var foundAlbum = none[Album]()

  # If we know the year the album was released, we can use that to narrow
  # down the list of matching albums
  if year.isSome:
    # First look only at albums already associated with this artist
    for artist in artists:
      let existing = lib.db.findAlbumsByArtistNameAndYear(artist, name, year.get)
      if existing.records.len > 0:
        foundAlbum = some(existing.records[0])
        break

    # Then look at all albums with that name and year
    if foundAlbum.isNone:
      let existing = lib.db.findAlbumsByNameAndYear(name, $year.get)
      if existing.records.len > 0: foundAlbum = some(existing.records[0])

  if foundAlbum.isNone:
    # If we don't know the year, or if there are no albums with that year,
    # look at all albums by this artist
    for artist in artists:
      let existing = lib.db.findAlbumsByArtistAndName(artist, name)
      if existing.records.len > 0:
        foundAlbum = some(existing.records[0])
        break

  # If we still don't have a match, look at all albums only by name
  if foundAlbum.isNone:
    let existing = lib.db.findAlbumsByName(name)
    if existing.records.len > 0: foundAlbum = some(existing.records[0])

  # If we still don't have a match, create a new album
  if foundAlbum.isNone:
    getLogger("wdiwtlt/medialibrary").debug(%*{
      "msg": "Creating missing Album record.",
      "method": "findOrCreateAlbum",
      "albumName": name,
      "year": year.map((y) => $y).get("") })

    foundAlbum = some(lib.db.createAlbum(Album(
      id: genUuid(),
      name: name,
      year: year,
      trackTotal: 0)))

  return foundAlbum.get

proc associateWithAristsAndAlbums*(
    lib: MediaLibrary,
    mf: MediaFile,
    artistNames: seq[string],
    albumNames: seq[string],
    year: Option[int]) =

  # Find or create artist and album records.
  let artists = artistNames.mapIt(lib.findOrCreateArtist(it))

  # Find or create album records.
  let albums = albumNames.mapIt(lib.findOrCreateAlbum(it, artists, year))

  # Associate this file with the artists and albums.
  for artist in artists: lib.db.associate(artist, mf)
  for album in albums: lib.db.associate(album, mf)

  # Make sure we have associations between all artists and albums.
  for artist in artists:
    let albumsForArtist = lib.db.findAlbumsByArtistId($artist.id)
    for album in albums:
      if not albumsForArtist.records.anyIt(album.id == it.id):
        lib.db.associate(artist, album)

proc addFile*(lib: MediaLibrary, relativeFilePath: Path): MediaFile =

  let fullPath = lib.root / relativeFilePath
  let pathParts = splitFile(fullPath)

  if not fullPath.fileExists:
    raise newException(ValueError, "File does not exist: " & $fullPath)

  let existing = lib.db.findMediaFilesByFilePath($relativeFilePath)
  if existing.records.len > 0:
    getLogger("wdiwtlt/medialibrary").debug(%*{
      "msg": "File already exists in library, using existing record.",
      "method": "addFile",
      "relativeFilePath": $relativeFilePath,
      "existingId": existing.records[0].id})
    return existing.records[0]

  var newMf = MediaFile(
    id: genUuid(),
    playCount: 0,
    filePath: $relativeFilePath,
    dateAdded: now(),
    lastPlayed: none[DateTime](),
    presentLocally: true,
    comment: "")

  newMf.fileHash = fileToMD5(fullPath)

  let existingForHash = lib.db.findMediaFilesByFileHash(newMf.fileHash)
  if existingForHash.records.len > 0:
    getLogger("wdiwtlt/medialibrary").debug(%*{
      "msg": "File with the same hash already exists in library, using existing record.",
      "method": "addFile",
      "relativeFilePath": $relativeFilePath,
      "existingId": existingForHash.records[0].id})
    return existingForHash.records[0]

  var tagFile = openTags(fullPath)
  defer: close(tagFile)

  newMf.name =
    if tagFile.title.strip().len == 0: $pathParts.name
    else: tagFile.title.strip()

  newMf.comment = tagFile.comment.strip()
  newMf.discNumber = tagFile.discNumber
  newMf.trackNumber =
    if tagFile.track == 0: none[int]()
    else: some(tagFile.track)

  let folderParts = ($pathParts.dir).split("/")

  var artistNames = newSeq[string]()
  if tagFile.artist.strip().len > 0:
    newMf.metaInfoSource = "tag info"
    artistNames = tagFile.artist.split({'/', ';'}).mapIt(strip(it))
  elif tagFile.albumArtist.strip().len > 0:
    newMf.metaInfoSource = "tag info"
    artistNames = tagFile.albumArtist.split({'/', ';'}).mapIt(strip(it))
  elif folderParts.len > 1:
    newMf.metaInfoSource = "folder"
    artistNames = @[folderParts[folderParts.len - 2].strip()]
  else:
    newMf.metaInfoSource = "unknown"
    artistNames = @[]

  var albumNames = newSeq[string]()
  if tagFile.album.strip().len > 0:
    newMf.metaInfoSource = "tag info"
    albumNames = tagFile.album.split({'/', ';'}).mapIt(strip(it))
  elif folderParts.len > 0:
    newMf.metaInfoSource = "folder"
    albumNames = @[folderParts[folderParts.len - 1].strip()]
  else:
    newMf.metaInfoSource = "unknown"
    albumNames = @[]

  result = lib.db.createMediaFile(newMf)

  let albumYear =
    if tagFile.year == 0: none[int]()
    else: some(tagFile.year)

  lib.associateWithAristsAndAlbums(result, artistNames, albumNames, albumYear)

proc rescanLibrary*(lib: MediaLibrary):
    tuple[ total, ignored, new, present, absent: int] =

  var missingFiles = lib.db.getAllMediaFiles().records
  var foundFiles = newSeq[MediaFile]()

  let startDate = now()
  for p in lib.root.walkDir(
      relative = true,
      checkDir = true,
      skipSpecial = true):

    if not p.path.fileExists: continue
    let mf = lib.addFile(p.path)
    result.total += 1

    foundFiles.add(mf)
    missingFiles = missingFiles.filterIt(it.id != mf.id)

    if mf.dateAdded > startDate:
      result.new += 1

  for mf in foundFiles:
    if not mf.presentLocally:
      var updated = mf
      updated.presentLocally = true
      discard lib.db.updateMediaFile(updated)

  for mf in missingFiles:
    if mf.presentLocally:
      var updated = mf
      updated.presentLocally = false
      discard lib.db.updateMediaFile(updated)

  result.present = foundFiles.len
  result.absent = missingFiles.len
